/*
 * kgraft_patch_bsc1014271
 *
 * Fix for CVE-2016-9576, bsc#1014271
 *
 *  SLE12(-SP1) commit:
 *  13e54ca18738fa8fa8eb09e1e67119569787532e
 *  	edited though
 *
 *  Copyright (c) 2016 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/pipe_fs_i.h>
#include <linux/splice.h>
#include <linux/uio.h>
#include <linux/slab.h>
#include <linux/pagemap.h>

static struct pipe_buf_operations *kgr_default_pipe_buf_ops;
static void (*kgr_spd_release_page)(struct splice_pipe_desc *spd, unsigned int i);
static int (*kgr_splice_grow_spd)(const struct pipe_inode_info *pipe, struct splice_pipe_desc *spd);
static ssize_t (*kgr_splice_to_pipe)(struct pipe_inode_info *pipe, struct splice_pipe_desc *spd);
static int (*kgr_write_pipe_buf)(struct pipe_inode_info *pipe, struct pipe_buffer *buf, struct splice_desc *sd);
static ssize_t (*kgr_splice_from_pipe)(struct pipe_inode_info *pipe, struct file *out, loff_t *ppos, size_t len, unsigned int flags, splice_actor *actor);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "default_pipe_buf_ops", (void *)&kgr_default_pipe_buf_ops },
	{ "spd_release_page", (void *)&kgr_spd_release_page },
	{ "splice_grow_spd", (void *)&kgr_splice_grow_spd },
	{ "splice_to_pipe", (void *)&kgr_splice_to_pipe },
	{ "write_pipe_buf", (void *)&kgr_write_pipe_buf },
	{ "splice_from_pipe", (void *)&kgr_splice_from_pipe },
};

/* inlined */
static ssize_t kgr_kernel_readv(struct file *file, const struct iovec *vec,
			    unsigned long vlen, loff_t offset)
{
	mm_segment_t old_fs;
	loff_t pos = offset;
	ssize_t res;

	old_fs = get_fs();
	set_fs(get_ds());
	/* The cast to a user pointer is valid due to the set_fs() */
	res = vfs_readv(file, (const struct iovec __user *)vec, vlen, &pos);
	set_fs(old_fs);

	return res;
}

/* inlined */
static void kgr_splice_shrink_spd(struct splice_pipe_desc *spd)
{
	if (spd->nr_pages_max <= PIPE_DEF_BUFFERS)
		return;

	kfree(spd->pages);
	kfree(spd->partial);
}

/* patched function */
ssize_t kgr_default_file_splice_read(struct file *in, loff_t *ppos,
	struct pipe_inode_info *pipe, size_t len, unsigned int flags)
{
	unsigned int nr_pages;
	unsigned int nr_freed;
	size_t offset;
	struct page *pages[PIPE_DEF_BUFFERS];
	struct partial_page partial[PIPE_DEF_BUFFERS];
	struct iovec *vec, __vec[PIPE_DEF_BUFFERS];
	ssize_t res;
	size_t this_len;
	int error;
	int i;
	struct splice_pipe_desc spd = {
		.pages = pages,
		.partial = partial,
		.nr_pages_max = PIPE_DEF_BUFFERS,
		.flags = flags,
		.ops = kgr_default_pipe_buf_ops,
		.spd_release = kgr_spd_release_page,
	};

	/*
	 * Fix CVE-2016-9576
	 *  +3 lines
	 */
	if (!S_ISREG(file_inode(in)->i_mode))
		return -EINVAL;

	if (kgr_splice_grow_spd(pipe, &spd))
		return -ENOMEM;

	res = -ENOMEM;
	vec = __vec;
	if (spd.nr_pages_max > PIPE_DEF_BUFFERS) {
		vec = kmalloc(spd.nr_pages_max * sizeof(struct iovec), GFP_KERNEL);
		if (!vec)
			goto shrink_ret;
	}

	offset = *ppos & ~PAGE_CACHE_MASK;
	nr_pages = (len + offset + PAGE_CACHE_SIZE - 1) >> PAGE_CACHE_SHIFT;

	for (i = 0; i < nr_pages && i < spd.nr_pages_max && len; i++) {
		struct page *page;

		page = alloc_page(GFP_USER);
		error = -ENOMEM;
		if (!page)
			goto err;

		this_len = min_t(size_t, len, PAGE_CACHE_SIZE - offset);
		vec[i].iov_base = (void __user *) page_address(page);
		vec[i].iov_len = this_len;
		spd.pages[i] = page;
		spd.nr_pages++;
		len -= this_len;
		offset = 0;
	}

	res = kgr_kernel_readv(in, vec, spd.nr_pages, *ppos);
	if (res < 0) {
		error = res;
		goto err;
	}

	error = 0;
	if (!res)
		goto err;

	nr_freed = 0;
	for (i = 0; i < spd.nr_pages; i++) {
		this_len = min_t(size_t, vec[i].iov_len, res);
		spd.partial[i].offset = 0;
		spd.partial[i].len = this_len;
		if (!this_len) {
			__free_page(spd.pages[i]);
			spd.pages[i] = NULL;
			nr_freed++;
		}
		res -= this_len;
	}
	spd.nr_pages -= nr_freed;

	res = kgr_splice_to_pipe(pipe, &spd);
	if (res > 0)
		*ppos += res;

shrink_ret:
	if (vec != __vec)
		kfree(vec);
	kgr_splice_shrink_spd(&spd);
	return res;

err:
	for (i = 0; i < spd.nr_pages; i++)
		__free_page(spd.pages[i]);

	res = error;
	goto shrink_ret;
}

/* patched function */
ssize_t kgr_default_file_splice_write(struct pipe_inode_info *pipe,
	struct file *out, loff_t *ppos, size_t len, unsigned int flags)
{
	ssize_t ret;

	/*
	 * Fix CVE-2016-9576
	 *  +3 lines
	 */
	if (!S_ISREG(file_inode(out)->i_mode))
		return -EINVAL;

	ret = kgr_splice_from_pipe(pipe, out, ppos, len, flags, kgr_write_pipe_buf);
	if (ret > 0)
		*ppos += ret;

	return ret;
}

static int kgr_patch_bsc1014271_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

int kgr_patch_bsc1014271_init(void)
{
	return kgr_patch_bsc1014271_kallsyms();
}
